# wwwpy/pytest_plugin.py
from pathlib import Path
import pytest
import importlib.util
import inspect

from playwright.sync_api import Page

from wwwpy.server.pytest.playwright import playwright_setup_page_logger
from wwwpy.server.pytest.xvirt_impl import XVirtImpl
import wwwpy.server.pytest.playwright

def pytest_addoption(parser):
    parser.addoption("--headful", action="store_true", default=False, help="run tests in headfull mode")


def pytest_configure(config):
    # This function is called before any tests or fixtures are collected.
    # You can add any setup code here.

    # For example, you can add your conftest.py setup code here.
    # Note: You might need to adjust this code to work in this context.
    from wwwpy.server.pytest.playwright import playwright_patch_timeout
    playwright_patch_timeout()

def _get_package_path(package_name: str) -> Path:
    spec = importlib.util.find_spec(package_name)
    if spec:
        package = importlib.import_module(package_name)
        package_path = inspect.getfile(package)
        return Path(package_path)
    return None

@pytest.hookimpl
def pytest_xvirt_setup(config):
    package_name = 'tests.remote'

    location = _get_package_path(package_name)
    folder_remote = location.parent
    folder_root = folder_remote.parent.parent
    headfull = config.getoption("--headful")
    # Now you can use the headfull variable in your setup
    return XVirtImpl(folder_remote, folder_root, headless=not headfull)


def pytest_unconfigure(config):
    # This function is called after all tests and fixtures have been collected.
    # You can add any teardown code here.
    pass


@pytest.fixture(scope="function", autouse=True)
def before_each_after_each(page: Page):
    playwright_setup_page_logger(page)
    yield
    # print("I just experienced that this is not printed if the test fails")



from __future__ import annotations

from pathlib import Path
from typing import List

from wwwpy.bootstrap import bootstrap_routes
from wwwpy.resources import library_resources, from_directory, from_file, \
    FilesystemIterable, ResourceIterable
from wwwpy.rpc import Services, Module, Stubber
from wwwpy.webserver import wait_forever
from wwwpy.webservers.available_webservers import available_webservers


def start_default(port: int, directory: Path):
    webserver = available_webservers().new_instance()

    _convention(webserver, directory)

    webserver.set_port(port).start_listen()
    wait_forever()


def _add_conventional_resources(resources: List[ResourceIterable], directory: Path, relative_to: Path = None):
    if relative_to is None:
        relative_to = directory

    resources.extend([
        from_directory(directory / 'remote', relative_to=relative_to),
        from_directory(directory / 'common', relative_to=relative_to),
        from_file(directory / 'common.py', relative_to=relative_to),
        from_file(directory / 'remote.py', relative_to=relative_to),
        *_conventional_resources_additional
    ])


def _convention(webserver, directory) -> List[ResourceIterable]:
    """
    Convention for a wwwpy server.
    It configures the webserver to serve the files from the working directory.
    It also configures the webserver to serve the files from the library.
    """
    print(f'applying convention to working_dir: {directory}')
    import sys
    sys.path.insert(0, str(directory))
    resources = []
    try:
        import server.rpc as rpc_module
        services = Services()
        rpc_module = Module(rpc_module)
        services.add_module(rpc_module)
        webserver.set_http_route(services.route)

        resources.append(Stubber(services.route.path, services, rpc_module).remote_stub_resources())
    except Exception as e:
        print(f'could not load rpc module: {e}')

    _add_conventional_resources(resources, directory)

    resources.append(library_resources())
    bootstrap_python = f'from wwwpy.remote.main import entry_point; await entry_point()'
    webserver.set_http_route(*bootstrap_routes(resources, python=bootstrap_python))

    return resources


def _setup_default_bootrap(resources, webserver):
    resources.append(library_resources())
    bootstrap_python = f'from wwwpy.remote.main import entry_point; await entry_point()'
    webserver.set_http_route(*bootstrap_routes(resources, python=bootstrap_python))


_conventional_resources_additional: List[FilesystemIterable] = []


def _conventional_resources_additional_append(filesystem_iterable: FilesystemIterable):
    _conventional_resources_additional.append(filesystem_iterable)



import os


async def install(package):
    from js import pyodide
    await pyodide.loadPackage('micropip')
    import micropip
    await micropip.install([package])


async def main(rootpath, invocation_dir, args):
    await install('pytest==7.2.2')  # didn't work with update to 8.1.1
    await install('pytest-xvirt')
    import pytest
    print('-=-' * 20 + 'pytest imported')

    from wwwpy.common.tree import print_tree
    print_tree('/wwwpy_bundle')

    os.chdir(invocation_dir)
    pytest.main(args)



import io

import asyncio

class HttpStreaming:
    """request method, e.g., GET, POST, etc."""
    method: str
    """request content type, e.g., application/json, text/html, etc."""
    content_type: str

    def stream(self, content_type: str) -> io.RawIOBase:
        """Returns a stream to write the response content.
          Args:
                content_type (str): the content type of the response.
        """
        raise NotImplementedError

class SansIOProtocol:
    """A protocol that does not use IO."""
    def process(self, data: bytes):
        """Processes the data."""
        raise NotImplementedError



import json
import threading
from pathlib import Path
from queue import Queue
from threading import Thread

from playwright.sync_api import sync_playwright
from xvirt import XVirt

from wwwpy.server.pytest.playwright import playwright_setup_page_logger, start_playwright, start_playwright_in_thread
from wwwpy.bootstrap import bootstrap_routes
from wwwpy.http import HttpRoute, HttpRequest, HttpResponse
from wwwpy.resources import library_resources, from_directory, StringResource
from wwwpy.server import find_port
from wwwpy.webservers.python_embedded import WsPythonEmbedded

_file_parent = Path(__file__).parent


class XVirtImpl(XVirt):

    def __init__(self, parent_remote: Path, relative_to: Path, headless: bool = True):
        self.parent_remote = parent_remote
        self.relative_to = relative_to
        self.headless = headless

        self.events = Queue()
        self.close_pw = threading.Event()

    def virtual_path(self) -> str:
        return str(self.parent_remote)

    def _http_handler(self, req: HttpRequest) -> HttpResponse:
        print(f'server side xvirt_notify_handler({req})')
        self.events.put(req.content)
        return HttpResponse('', 'text/plain')

    def run(self):
        webserver = self._start_webserver()
        start_playwright_in_thread(webserver.localhost_url(), self.headless)

    def _start_webserver(self):
        xvirt_notify_route = HttpRoute('/xvirt_notify', self._http_handler)
        # read remote conftest content
        remote_conftest = (_file_parent / 'remote_conftest.py').read_text() \
            .replace('#xvirt_notify_path_marker#', '/xvirt_notify')

        resources = [library_resources(),
                     from_directory(self.parent_remote, relative_to=self.relative_to),
                     [
                         # StringResource('tests/__init__.py', ''),
                         # StringResource('pytest.ini', ''),
                         StringResource('conftest.py', remote_conftest),
                         StringResource('remote_test_main.py',
                                        (_file_parent / 'remote_test_main.py').read_text())],

                     ]
        webserver = WsPythonEmbedded()
        invocation_dir, args = self.remote_invocation_params('/wwwpy_bundle')
        invocation_dir_json = json.dumps(invocation_dir)
        args_json = json.dumps(args)
        rootpath = json.dumps('/wwwpy_bundle')
        bootstrap_python = f'import remote_test_main; await remote_test_main.main({rootpath},{invocation_dir_json},{args_json})'
        webserver.set_http_route(*bootstrap_routes(resources, python=bootstrap_python), xvirt_notify_route)
        webserver.set_port(find_port()).start_listen()
        return webserver

    def finalize(self):
        self.close_pw.set()
        # self._thread.join()

    def recv_event(self) -> str:
        return self.events.get(timeout=30)
